/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package eu.pedu.b62_j14.c26_records;

import java.util.function.DoubleToIntFunction;

/**
 *
 * @author  Rudolf PECINOVSKÝ
 */

/***************************************************************************
 * Pravoúhlý trojúhelník definovaný délkami stran podle věty SSS.
 */
public record RightTriangleSSS(double a, double b, double c, int ID)
implements ITriangleSSS
{
    private static final int MI = Integer.MAX_VALUE;
    private static int count = 0;

    /***********************************************************************
     * Vytvoří záznam o pravoúhlém trojúhelníku zadaném délkami svých stran.
     * Kontroluje platnost Pythagorovy věty.
     */
    public RightTriangleSSS
    {
        if (ID != MI) {
            throw new IllegalArgumentException(
                "\nPoužívání konstruktoru zadávajícího všechny délky stran"
              + " je zakázáno!");
        }
        this.ID = ++count;
    }

    /***********************************************************************
     * Vytvoří záznam o pravoúhlém trojúhelníku zadaném délkou odvěsen.
     * Délku přepony dopočte podle Pythagorovy věty.
     */
    public RightTriangleSSS(double a, double b)
    {
        this(a, b, Math.sqrt(a*a + b*b), MI);
    }

    /***********************************************************************
     * Vytvoří záznam o rovnoramenném pravoúhlém trojúhelníku zadaném délkou
     * své přepony. Délku odvěsen dopočte podle Pythagorovy věty.
     */
    public RightTriangleSSS(double c)
    {
        this(Math.sqrt(c*c / 2), Math.sqrt(c*c / 2), c, MI);
    }

    /***********************************************************************
     * Spočte úhly u přepony pravoúhlého trojúhelníka a vrátí pole
     * s jejich velikostí zadanou ve stupních.
     */
    public double[] alphaBeta()
    {
        double alpha = Math.toDegrees(Math.acos(a()/c()));
        double beta  = Math.toDegrees(Math.acos(b()/c()));
        return new double[] { alpha, beta };
    }

    /***********************************************************************
     * Při porovnávání instancí kontroluje pouze délky stran;
     * hodnota ID ani konkrétní typ objektu jej nezajímají.
     */
    public boolean equals(Object o)
    {
        var result = false;
        if (o instanceof ITriangleSSS t) {
            result = (a == t.a())  &&  (b == t.b())  &&  (c==t.c());
        }
        return result;
    }

    /***********************************************************************
     * Generuje heškód závisející pouze na délkách stran.
     */
    public int hashCode()
    {
        DoubleToIntFunction h = (double d) ->
                           (int)( Double.doubleToLongBits(d) ^
                                 (Double.doubleToLongBits(d) >>> 32));
        int hash = ((97 + h.applyAsInt(a)) *97
                        + h.applyAsInt(b) )*97 + h.applyAsInt(c);
        return hash;
    }

}
